import pandas as pd
import numpy as np
import datetime
from typing import Tuple

from . import data_spacescope, data_starboard
DEFAULT_DATA_BACKEND = 'spacescope'

spacescope_obj = None
def setup_spacescope(auth_config_or_token):
    global spacescope_obj
    spacescope_obj = data_spacescope.SpacescopeDataConnection(auth_config_or_token)

def check_if_spacescope_configured():
    if spacescope_obj is None:
        raise ValueError("Spacescope object not configured. Please run setup_spacescope() with an bearer token configuration first!")

def get_historical_network_stats(
    start_date: datetime.date, 
    current_date: datetime.date, 
    end_date: datetime.date,
    data_backend: str = DEFAULT_DATA_BACKEND
) -> pd.DataFrame:
    if data_backend.lower() == 'spacescope':
        check_if_spacescope_configured()
        return spacescope_obj.get_historical_network_stats(start_date, current_date, end_date)
    elif data_backend.lower() == 'starboard':
        return data_starboard.get_historical_network_stats(start_date, current_date, end_date)
    else:
        raise ValueError("Data Backend: %s not supported!" % (data_backend,))

def get_sector_expiration_stats(
    start_date: datetime.date,
    current_date: datetime.date,
    end_date: datetime.date,
    data_backend: str = DEFAULT_DATA_BACKEND
) -> pd.DataFrame:
    if data_backend.lower() == 'spacescope':
        check_if_spacescope_configured()
        return spacescope_obj.get_sector_expiration_stats(start_date, current_date, end_date)
    elif data_backend.lower() == 'starboard':
        return data_starboard.get_sector_expiration_stats(start_date, current_date, end_date)
    else:
        raise ValueError("Data Backend: %s not supported!" % (data_backend,))

def get_day_renewed_power_stats(
    start_date: datetime.date,
    current_date: datetime.date,
    end_date: datetime.date,
    data_backend: str = DEFAULT_DATA_BACKEND
) -> Tuple[np.array, np.array]:
    if data_backend.lower() == 'spacescope':
        check_if_spacescope_configured()
        return spacescope_obj.get_day_renewed_power_stats(start_date, current_date, end_date)
    elif data_backend.lower() == 'starboard':
        return data_starboard.get_day_renewed_power_stats(start_date, current_date, end_date)
    else:
        raise ValueError("Data Backend: %s not supported!" % (data_backend,))

def query_sector_expirations(
    start_date: datetime.date, 
    end_date: datetime.date,
    data_backend: str = DEFAULT_DATA_BACKEND
) -> pd.DataFrame:
    if data_backend.lower() == 'spacescope':
        check_if_spacescope_configured()
        return spacescope_obj.query_spacescope_sector_expirations(start_date, end_date)
    elif data_backend.lower() == 'starboard':
        return data_starboard.query_starboard_sector_expirations(start_date, end_date)
    else:
        raise ValueError("Data Backend: %s not supported!" % (data_backend,))


def query_supply_stats(
    start_date: datetime.date,
    end_date: datetime.date,
    data_backend: str = DEFAULT_DATA_BACKEND
) -> pd.DataFrame:
    if data_backend.lower() == 'spacescope':
        check_if_spacescope_configured()
        return spacescope_obj.query_spacescope_supply_stats(start_date, end_date)
    elif data_backend.lower() == 'starboard':
        return data_starboard.query_starboard_supply_stats(start_date, end_date)
    else:
        raise ValueError("Data Backend: %s not supported!" % (data_backend,))


def query_power_stats(
    start_date: datetime.date, 
    end_date: datetime.date,
    data_backend: str = DEFAULT_DATA_BACKEND
) -> pd.DataFrame:
    if data_backend.lower() == 'spacescope':
        check_if_spacescope_configured()
        return spacescope_obj.query_spacescope_power_stats(start_date, end_date)
    elif data_backend.lower() == 'starboard':
        return data_starboard.query_starboard_power_stats(start_date, end_date)
    else:
        raise ValueError("Data Backend: %s not supported!" % (data_backend,))


def query_daily_power_onboarded(
    start_date: datetime.date,
    end_date: datetime.date,
    data_backend: str = DEFAULT_DATA_BACKEND
) -> pd.DataFrame:
    if data_backend.lower() == 'spacescope':
        check_if_spacescope_configured()
        return spacescope_obj.query_spacescope_daily_power_onboarded(start_date, end_date)
    elif data_backend.lower() == 'starboard':
        return data_starboard.query_starboard_daily_power_onboarded(start_date, end_date)
    else:
        raise ValueError("Data Backend: %s not supported!" % (data_backend,))


def query_supply_stats(
    start_date: datetime.date,
    end_date: datetime.date,
    data_backend: str = DEFAULT_DATA_BACKEND
) -> pd.DataFrame:
    if data_backend.lower() == 'spacescope':
        check_if_spacescope_configured()
        return spacescope_obj.query_spacescope_supply_stats(start_date, end_date)
    elif data_backend.lower() == 'starboard':
        return data_starboard.query_starboard_supply_stats(start_date, end_date)
    else:
        raise ValueError("Data Backend: %s not supported!" % (data_backend,))


def query_sector_expirations(
start_date: datetime.date,
    end_date: datetime.date,
    data_backend: str = DEFAULT_DATA_BACKEND
) -> pd.DataFrame:
    if data_backend.lower() == 'spacescope':
        check_if_spacescope_configured()
        return spacescope_obj.query_spacescope_sector_expirations(start_date, end_date)
    elif data_backend.lower() == 'starboard':
        return data_starboard.query_starboard_sector_expirations(start_date, end_date)
    else:
        raise ValueError("Data Backend: %s not supported!" % (data_backend,))
    

def get_storage_baseline_value(
    date: datetime.date,
    data_backend: str = DEFAULT_DATA_BACKEND
) -> float:
    # Get baseline values from Starboard API
    bp_df = query_historical_baseline_power(data_backend)
    # Extract baseline value at date
    init_baseline_bytes = bp_df[bp_df["date"] >= pd.to_datetime(date, utc="UTC")].iloc[
        0, 1
    ]
    return init_baseline_bytes
    

def get_cum_capped_rb_power(
    date: datetime.date,
    data_backend: str = DEFAULT_DATA_BACKEND
) -> float:
    # Query data sources and join
    rbp_df = query_historical_rb_power(data_backend)
    bp_df = query_historical_baseline_power(data_backend)
    df = pd.merge(rbp_df, bp_df, on="date", how="inner")
    # Compute cumulative capped RB power
    df["capped_power"] = np.min(df[["baseline", "rb_power"]].values, axis=1)
    df["cum_capped_power"] = df["capped_power"].cumsum()
    date_df = df[df["date"] >= pd.to_datetime(date, utc="UTC")]
    init_cum_capped_power = date_df["cum_capped_power"].iloc[0]
    return init_cum_capped_power


def get_cum_capped_qa_power(
    date: datetime.date,
    data_backend: str = DEFAULT_DATA_BACKEND
) -> float:
    # Query data sources and join
    qap_df = query_historical_qa_power(data_backend)
    bp_df = query_historical_baseline_power(data_backend)
    df = pd.merge(qap_df, bp_df, on="date", how="inner")
    # Compute cumulative capped RB power
    df["capped_power"] = np.min(df[["baseline", "qa_power"]].values, axis=1)
    df["cum_capped_power"] = df["capped_power"].cumsum()
    date_df = df[df["date"] >= pd.to_datetime(date, utc="UTC")]
    init_cum_capped_power = date_df["cum_capped_power"].iloc[0]
    return init_cum_capped_power


def get_vested_amount(
    date: datetime.date,
    data_backend: str = DEFAULT_DATA_BACKEND
) -> float:
    start_date = date - datetime.timedelta(days=1)
    end_date = date + datetime.timedelta(days=1)
    stats_df = query_supply_stats(start_date, end_date, data_backend)
    date_stats = stats_df[stats_df["date"] == date]
    return date_stats["vested_fil"].iloc[0]


def query_historical_baseline_power(
    data_backend: str = DEFAULT_DATA_BACKEND
) -> pd.DataFrame:
    if data_backend.lower() == 'spacescope':
        check_if_spacescope_configured()
        return spacescope_obj.query_historical_baseline_power()
    elif data_backend.lower() == 'starboard':
        return data_starboard.query_historical_baseline_power()
    else:
        raise ValueError("Data Backend: %s not supported!" % (data_backend,))


def query_historical_rb_power(
    data_backend: str = DEFAULT_DATA_BACKEND
) -> pd.DataFrame:
    if data_backend.lower() == 'spacescope':
        check_if_spacescope_configured()
        return spacescope_obj.query_historical_rb_power()
    elif data_backend.lower() == 'starboard':
        return data_starboard.query_historical_rb_power()
    else:
        raise ValueError("Data Backend: %s not supported!" % (data_backend,))


def query_historical_qa_power(
    data_backend: str = DEFAULT_DATA_BACKEND
) -> pd.DataFrame:
    if data_backend.lower() == 'spacescope':
        check_if_spacescope_configured()
        return spacescope_obj.query_historical_qa_power()
    elif data_backend.lower() == 'starboard':
        return data_starboard.query_historical_qa_power()
    else:
        raise ValueError("Data Backend: %s not supported!" % (data_backend,))



"""MechaFIL: Mechanistic model for the Filecoin Economy"""

__version__ = "1.7"
__author__ = "Maria Silva <misilva73@gmail.com>, Tom Mellan <t.mellan@imperial.ac.uk>, Kiran Karra <kiran.karra@gmail.com>"
__all__ = []

import pandas

pandas.set_option("mode.chained_assignment", None)
from .sim import run_simple_sim



import numpy as np
import pandas as pd
import datetime

from .data import get_vested_amount

FIL_BASE = 2_000_000_000.0
PL_AMOUNT = 0.15 * FIL_BASE
FOUNDATION_AMOUNT = 0.05 * FIL_BASE
STORAGE_MINING = 0.55 * FIL_BASE
MINING_RESERVE = 0.15 * FIL_BASE


def compute_vesting_trajectory_df(
    start_date: datetime.date, end_date: datetime.date
) -> pd.DataFrame:
    """
    15% to PL -> 6-year linear vesting
    5% to FIlecoin foundation -> 6-year linear vesting
    10% to Investors -> Linear vesting with different durations (taken from lotus):
        - 0 days: 10_632_000
        - 6 months: 19_015_887 + 32_787_700
        - 1 yrs: 22_421_712 + 9_400_000
        - 2 yrs: 7_223_364
        - 3 yrs: 87_637_883 + 898_958
        - 6 yrs: 9_805_053
        (total of 199_822_557)

    Info taken from:
        - https://coinlist.co/assets/index/filecoin_2017_index/Filecoin-Sale-Economics-e3f703f8cd5f644aecd7ae3860ce932064ce014dd60de115d67ff1e9047ffa8e.pdf
        - https://spec.filecoin.io/#section-systems.filecoin_token.token_allocation
        - https://filecoin.io/blog/filecoin-circulating-supply/
        - https://github.com/filecoin-project/lotus/blob/e65fae28de2a44947dd24af8c7dafcade29af1a4/chain/stmgr/supply.go#L148
    """
    # we assume vesting started at main net launch, in 2020-10-15
    launch_date = datetime.date(2020, 10, 15)
    end_day = (end_date - launch_date).days
    # Get entire daily vesting trajectory
    full_vest_df = pd.DataFrame(
        {
            "date": pd.date_range(launch_date, end_date, freq="d")[:-1],
            "six_month_vest_saft": vest(19_015_887 + 32_787_700, 183, end_day),
            "one_year_vest_saft": vest(22_421_712 + 9_400_000, 365 * 1, end_day),
            "two_year_vest_saft": vest(7_223_364, 365 * 2, end_day),
            "three_year_vest_saft": vest(87_637_883 + 898_958, 365 * 3, end_day),
            "six_year_vest_saft": vest(9_805_053, 365 * 6, end_day),
            "six_year_vest_pl": vest(PL_AMOUNT, 365 * 6, end_day),
            "six_year_vest_foundation": vest(FOUNDATION_AMOUNT, 365 * 6, end_day),
        }
    )
    full_vest_df["date"] = full_vest_df["date"].dt.date
    # Filter vesting trajectory for desired dates
    vest_df = full_vest_df[full_vest_df["date"] >= start_date]
    # Compute total cumulative vesting
    vest_df.loc[:, "total_day_vest"] = vest_df.drop(columns=["date"]).sum(axis=1)
    start_vested_amt = get_vested_amount(start_date)
    vest_df.loc[:, "total_vest"] = vest_df["total_day_vest"].cumsum() + start_vested_amt
    vest_df = vest_df[["date", "total_vest"]]
    return vest_df


def vest(amount: float, time: int, end_day: int) -> np.array:
    """
    amount -- total amount e.g 300M FIL for SAFT
    time -- vesting time in days
    end_day -- end day for the vesting trajectory
    """
    ones_ = np.ones(int(time))[:end_day]
    extra_to_pad_ = max(0, end_day - int(time))
    ones_padded_ = np.pad(ones_, (0, extra_to_pad_))
    vest_ = ones_padded_ / time
    return amount * vest_



import requests
import pandas as pd
import numpy as np
import datetime
from typing import Tuple

EXBI = 2**60
PIB = 2**50


def get_historical_network_stats(
    start_date: datetime.date, current_date: datetime.date, end_date: datetime.date
) -> pd.DataFrame:
    power_df = query_starboard_power_stats(start_date, current_date)
    onboards_df = query_starboard_daily_power_onboarded(start_date, current_date)
    stats_df = query_starboard_supply_stats(start_date, current_date)
    stats_df = stats_df.merge(power_df, on="date", how="inner").merge(
        onboards_df, on="date", how="inner"
    )
    rb_renewal_rate, day_renewed_qa_power = get_day_renewed_power_stats(
        start_date, current_date, end_date
    )
    stats_df["rb_renewal_rate"] = rb_renewal_rate
    stats_df["day_renewed_qa_power_pib"] = day_renewed_qa_power
    return stats_df


def get_sector_expiration_stats(
    start_date: datetime.date,
    current_date: datetime.date,
    end_date: datetime.date,
) -> pd.DataFrame:
    scheduled_df = query_starboard_sector_expirations(start_date, end_date)
    filter_scheduled_df = scheduled_df[
        scheduled_df["date"] >= pd.to_datetime(current_date, utc="UTC")
    ]
    rbp_expire_vec = filter_scheduled_df["total_rb"].values
    qap_expire_vec = filter_scheduled_df["total_qa"].values
    # we need the entire history of known_scheduled_pledge_release, so get the
    # data from the entire time-window, not just from current-date onwards
    pledge_release_vec = scheduled_df["total_pledge"].values
    return rbp_expire_vec, qap_expire_vec, pledge_release_vec


def get_day_renewed_power_stats(
    start_date: datetime.date,
    current_date: datetime.date,
    end_date: datetime.date,
) -> Tuple[np.array, np.array]:
    scheduled_df = query_starboard_sector_expirations(start_date, end_date)
    filter_scheduled_df = scheduled_df[
        scheduled_df["date"] < pd.to_datetime(current_date, utc="UTC")
    ]
    rb_renewal_rate = (
        filter_scheduled_df["extended_rb"] / filter_scheduled_df["total_rb"]
    ).values
    day_renewed_qa_power = filter_scheduled_df["extended_qa"].values
    return rb_renewal_rate, day_renewed_qa_power


def query_starboard_sector_expirations(
    start_date: datetime.date, end_date: datetime.date
) -> pd.DataFrame:
    url = f"https://observable-api.starboard.ventures/getdata/sectors_schedule_expiration_full?start={str(start_date)}&end={str(end_date)}"
    r = requests.get(url)
    # Put data in dataframe
    scheduled_df = pd.DataFrame(r.json()["data"])
    # Convert bytes to pebibytes
    scheduled_df["extended_rb"] = scheduled_df["extended_bytes"].astype(float) / PIB
    scheduled_df["expired_rb"] = scheduled_df["expired_bytes"].astype(float) / PIB
    scheduled_df["open_rb"] = scheduled_df["potential_expire_bytes"].astype(float) / PIB
    scheduled_df["extended_qa"] = scheduled_df["extended_bytes_qap"].astype(float) / PIB
    scheduled_df["expired_qa"] = scheduled_df["expired_bytes_qap"].astype(float) / PIB
    scheduled_df["open_qa"] = (
        scheduled_df["potential_expire_bytes_qap"].astype(float) / PIB
    )
    # Total scheduled to expire, excluding terminated
    scheduled_df["total_rb"] = (
        scheduled_df["extended_rb"]
        + scheduled_df["expired_rb"]
        + scheduled_df["open_rb"]
    )
    scheduled_df["total_qa"] = (
        scheduled_df["extended_qa"]
        + scheduled_df["expired_qa"]
        + scheduled_df["open_qa"]
    )
    scheduled_df["total_pledge"] = (
        scheduled_df["extended_pledge"].astype(float)
        + scheduled_df["expired_pledge"].astype(float)
        + scheduled_df["potential_expire_pledge"].astype(float)
    )

    scheduled_df["schedule_expire_rb"] = scheduled_df["schedule_expire_bytes"].astype(float) / PIB
    scheduled_df["schedule_expire_qa"] = scheduled_df["schedule_expire_bytes_qap"].astype(float) / PIB
    scheduled_df["schedule_expire_pledge"] = scheduled_df["schedule_expire_pledge"].astype(float)

    # Convert interest date to datetime
    scheduled_df["date"] = pd.to_datetime(scheduled_df["interest_date"])
    # Filter dates
    scheduled_df = scheduled_df[
        scheduled_df["date"] >= pd.to_datetime(start_date, utc="UTC")
    ]
    scheduled_df = scheduled_df[
        scheduled_df["date"] < pd.to_datetime(end_date, utc="UTC")
    ]
    return scheduled_df


def query_starboard_daily_power_onboarded(
    start_date: datetime.date,
    end_date: datetime.date,
) -> pd.DataFrame:
    # Get data from prove-commit-split-d API
    url = f"https://observable-api.starboard.ventures/api/v1/observable/prove-commit-split-d-v2?start={str(start_date)}&end={str(end_date)}"
    r = requests.get(url)
    onboards_df = pd.DataFrame(r.json()["data"])
    # Compute total onboardings
    onboards_df["day_onboarded_rb_power_pib"] = (
        onboards_df["half_size_byte"].astype(float)
        + onboards_df["size_byte"].astype(float)
    ) / PIB
    onboards_df["day_onboarded_qa_power_pib"] = (
        onboards_df["half_size_byte_qap"].astype(float)
        + onboards_df["size_byte_qap"].astype(float)
    ) / PIB
    # Convert dates to datetime
    onboards_df["date"] = pd.to_datetime(onboards_df["stat_date"]).dt.date
    # Filter dates
    onboards_df = onboards_df[
        (onboards_df["date"] >= start_date) & (onboards_df["date"] <= end_date)
    ]
    # Filter columns
    onboards_df = onboards_df[
        ["date", "day_onboarded_rb_power_pib", "day_onboarded_qa_power_pib"]
    ]
    return onboards_df


def query_starboard_supply_stats(
    start_date: datetime.date,
    end_date: datetime.date,
) -> pd.DataFrame:
    url = f"https://observable-api.starboard.ventures/api/v1/observable/circulating-supply?start={str(start_date)}&end={str(end_date)}"
    r = requests.get(url)
    raw_stats_df = pd.DataFrame(r.json()["data"])
    # Convert metrics to float
    stats_df = raw_stats_df[
        [
            "circulating_fil",
            "mined_fil",
            "vested_fil",
            "locked_fil",
            "burnt_fil",
        ]
    ].astype(float)
    # Convert dates to datetime dates
    stats_df["date"] = pd.to_datetime(raw_stats_df["stat_date"]).dt.date
    # Filter dates
    stats_df = stats_df[
        (stats_df["date"] >= start_date) & (stats_df["date"] <= end_date)
    ]
    return stats_df


def query_starboard_power_stats(
    start_date: datetime.date, end_date: datetime.date
) -> pd.DataFrame:
    url = f"https://observable-api.starboard.ventures/network_storage_capacity?start={str(start_date)}&end={str(end_date)}"
    r = requests.get(url)
    power_df = pd.DataFrame(r.json()["data"])
    # Convert dates to datetime dates
    power_df["date"] = pd.to_datetime(power_df["stat_date"]).dt.date
    # Filter dates
    power_df = power_df[
        (power_df["date"] >= start_date) & (power_df["date"] <= end_date)
    ]
    # Convert power stats to exibytes
    power_df["total_raw_power_eib"] = (
        power_df["total_raw_bytes_power"].astype(float) / EXBI
    )
    power_df["total_qa_power_eib"] = (
        power_df["total_qa_bytes_power"].astype(float) / EXBI
    )
    # Select final columns
    power_df = power_df[["date", "total_raw_power_eib", "total_qa_power_eib"]]
    return power_df


def query_historical_baseline_power() -> pd.DataFrame:
    url = f"https://observable-api.starboard.ventures/api/v1/observable/network-storage-capacity/new_baseline_power"
    r = requests.get(url)
    bp_df = pd.DataFrame(r.json()["data"])
    bp_df["date"] = pd.to_datetime(bp_df["stat_date"])
    bp_df["baseline"] = bp_df["new_baseline_power"].astype(float)
    bp_df = bp_df[["date", "baseline"]]
    return bp_df


def query_historical_rb_power() -> pd.DataFrame:
    url = f"https://observable-api.starboard.ventures/network_storage_capacity/total_raw_bytes_power"
    r = requests.get(url)
    rbp_df = pd.DataFrame(r.json()["data"])
    rbp_df["date"] = pd.to_datetime(rbp_df["stat_date"])
    rbp_df["rb_power"] = rbp_df["total_raw_bytes_power"].astype(float)
    rbp_df = rbp_df[["date", "rb_power"]]
    return rbp_df


def query_historical_qa_power() -> pd.DataFrame:
    url = f"https://observable-api.starboard.ventures/network_storage_capacity/total_qa_bytes_power"
    r = requests.get(url)
    qap_df = pd.DataFrame(r.json()["data"])
    qap_df["date"] = pd.to_datetime(qap_df["stat_date"])
    qap_df["qa_power"] = qap_df["total_qa_bytes_power"].astype(float)
    qap_df = qap_df[["date", "qa_power"]]
    return qap_df



def validate_qap_method(user_input):
    user_input = user_input.lower()
    if not (user_input == 'basic' or user_input == 'tunable'):
        raise ValueError("qap_method must be either basic or tunable!")


import numpy as np

GIB = 2**30

# Block reward collateral
def compute_day_locked_rewards(day_network_reward: float) -> float:
    return 0.75 * day_network_reward


def compute_day_reward_release(prev_network_locked_reward: float) -> float:
    return prev_network_locked_reward / 180.0


# Initial pledge collateral
def compute_day_delta_pledge(
    day_network_reward: float,
    prev_circ_supply: float,
    day_onboarded_qa_power: float,
    day_renewed_qa_power: float,
    total_qa_power: float,
    baseline_power: float,
    renewal_rate: float,
    scheduled_pledge_release: float,
    lock_target: float = 0.3,
) -> float:
    onboards_delta = compute_new_pledge_for_added_power(
        day_network_reward,
        prev_circ_supply,
        day_onboarded_qa_power,
        total_qa_power,
        baseline_power,
        lock_target,
    )
    renews_delta = compute_renewals_delta_pledge(
        day_network_reward,
        prev_circ_supply,
        day_renewed_qa_power,
        total_qa_power,
        baseline_power,
        renewal_rate,
        scheduled_pledge_release,
        lock_target,
    )
    return onboards_delta + renews_delta


def compute_day_locked_pledge(
    day_network_reward: float,
    prev_circ_supply: float,
    day_onboarded_qa_power: float,
    day_renewed_qa_power: float,
    total_qa_power: float,
    baseline_power: float,
    renewal_rate: float,
    scheduled_pledge_release: float,
    lock_target: float = 0.3,
) -> float:
    # Total locked from new onboards
    onboards_locked = compute_new_pledge_for_added_power(
        day_network_reward,
        prev_circ_supply,
        day_onboarded_qa_power,
        total_qa_power,
        baseline_power,
        lock_target,
    )
    # Total locked from renewals
    original_pledge = renewal_rate * scheduled_pledge_release
    new_pledge = compute_new_pledge_for_added_power(
        day_network_reward,
        prev_circ_supply,
        day_renewed_qa_power,
        total_qa_power,
        baseline_power,
        lock_target,
    )
    renews_locked = max(original_pledge, new_pledge)
    # Total locked pledge
    locked = onboards_locked + renews_locked
    return locked, renews_locked


def compute_renewals_delta_pledge(
    day_network_reward: float,
    prev_circ_supply: float,
    day_renewed_qa_power: float,
    total_qa_power: float,
    baseline_power: float,
    renewal_rate: float,
    scheduled_pledge_release: float,
    lock_target: float,
) -> float:
    # Delta from sectors expiring
    expire_delta = -(1 - renewal_rate) * scheduled_pledge_release
    # Delta from sector renewing
    original_pledge = renewal_rate * scheduled_pledge_release
    new_pledge = compute_new_pledge_for_added_power(
        day_network_reward,
        prev_circ_supply,
        day_renewed_qa_power,
        total_qa_power,
        baseline_power,
        lock_target,
    )
    renew_delta = max(0.0, new_pledge - original_pledge)
    # Delta for all scheduled sectors
    delta = expire_delta + renew_delta
    return delta


def compute_new_pledge_for_added_power(
    day_network_reward: float,
    prev_circ_supply: float,
    day_added_qa_power: float,
    total_qa_power: float,
    baseline_power: float,
    lock_target: float,
) -> float:
    # storage collateral
    storage_pledge = 20.0 * day_network_reward * (day_added_qa_power / total_qa_power)
    # consensus collateral
    normalized_qap_growth = day_added_qa_power / max(total_qa_power, baseline_power)
    consensus_pledge = max(lock_target * prev_circ_supply * normalized_qap_growth, 0)
    # total added pledge
    added_pledge = storage_pledge + consensus_pledge

    pledge_cap = day_added_qa_power * 1.0 / GIB  # The # of bytes in a GiB (Gibibyte)
    return min(pledge_cap, added_pledge)


def get_day_schedule_pledge_release(
    day_i,
    current_day_i,
    day_pledge_locked_vec: np.array,
    known_scheduled_pledge_release_vec: np.array,
    duration: int,
) -> float:
    # scheduled releases coming from known active sectors
    if day_i > len(known_scheduled_pledge_release_vec) - 1:
        known_day_release = 0.0
    else:
        known_day_release = known_scheduled_pledge_release_vec[day_i]
    # schedule releases coming from modeled sectors
    if day_i - duration >= current_day_i:
        model_day_release = day_pledge_locked_vec[day_i - duration]
    else:
        model_day_release = 0.0
    # Total pledge schedule releases
    day_pledge_schedules_release = known_day_release + model_day_release
    return day_pledge_schedules_release



import os

import pandas as pd
import numpy as np
import datetime
from typing import Union

from .data import get_historical_network_stats, get_sector_expiration_stats, setup_spacescope
from .power import (
    forecast_power_stats,
    build_full_power_stats_df,
    scalar_or_vector_to_vector,
)
from .vesting import compute_vesting_trajectory_df
from .minting import compute_minting_trajectory_df
from .supply import forecast_circulating_supply_df

from .utils import validate_qap_method

def setup_data_access(bearer_token_or_cfg: str):
    setup_spacescope(bearer_token_or_cfg)

def run_simple_sim(
    start_date: datetime.date,
    current_date: datetime.date,
    forecast_length: int,
    renewal_rate: Union[np.array, float],
    rb_onboard_power: Union[np.array, float],
    fil_plus_rate: Union[np.array, float],
    duration: int,
    bearer_token_or_cfg: str,
    qap_method: str = 'basic' # can be set to tunable or basic
                              # see: https://hackmd.io/O6HmAb--SgmxkjLWSpbN_A?view
) -> pd.DataFrame:
    validate_qap_method(qap_method)
    setup_data_access(bearer_token_or_cfg)

    end_date = current_date + datetime.timedelta(days=forecast_length)
    # Get sector scheduled expirations
    res = get_sector_expiration_stats(start_date, current_date, end_date)
    rb_known_scheduled_expire_vec = res[0]
    qa_known_scheduled_expire_vec = res[1]
    known_scheduled_pledge_release_full_vec = res[2]
    # Get daily stats
    fil_stats_df = get_historical_network_stats(start_date, current_date, end_date)
    current_day_stats = fil_stats_df[fil_stats_df["date"] >= current_date].iloc[0]
    # Forecast power stats
    rb_power_zero = current_day_stats["total_raw_power_eib"] * 1024.0
    qa_power_zero = current_day_stats["total_qa_power_eib"] * 1024.0
    rb_power_df, qa_power_df = forecast_power_stats(
        rb_power_zero,
        qa_power_zero,
        rb_onboard_power,
        rb_known_scheduled_expire_vec,
        qa_known_scheduled_expire_vec,
        renewal_rate,
        fil_plus_rate,
        duration,
        forecast_length,
        qap_method=qap_method
    )
    rb_power_df["total_raw_power_eib"] = rb_power_df["total_power"] / 1024.0
    qa_power_df["total_qa_power_eib"] = qa_power_df["total_power"] / 1024.0
    power_df = build_full_power_stats_df(
        fil_stats_df,
        rb_power_df,
        qa_power_df,
        start_date,
        current_date,
        end_date,
    )
    # Forecast Vesting
    vest_df = compute_vesting_trajectory_df(start_date, end_date)
    # Forecast minting stats and baseline
    rb_total_power_eib = power_df["total_raw_power_eib"].values
    qa_total_power_eib = power_df["total_qa_power_eib"].values
    qa_day_onboarded_power_pib = power_df["day_onboarded_qa_power_pib"].values
    qa_day_renewed_power_pib = power_df["day_renewed_qa_power_pib"].values
    mint_df = compute_minting_trajectory_df(
        start_date,
        end_date,
        rb_total_power_eib,
        qa_total_power_eib,
        qa_day_onboarded_power_pib,
        qa_day_renewed_power_pib,
    )
    # Forecast circulating supply
    start_day_stats = fil_stats_df.iloc[0]
    circ_supply_zero = start_day_stats["circulating_fil"]
    locked_fil_zero = start_day_stats["locked_fil"]
    daily_burnt_fil = fil_stats_df["burnt_fil"].diff().mean()
    burnt_fil_vec = fil_stats_df["burnt_fil"].values
    forecast_renewal_rate_vec = scalar_or_vector_to_vector(
        renewal_rate, forecast_length
    )
    past_renewal_rate_vec = fil_stats_df["rb_renewal_rate"].values[:-1]
    renewal_rate_vec = np.concatenate(
        [past_renewal_rate_vec, forecast_renewal_rate_vec]
    )
    cil_df = forecast_circulating_supply_df(
        start_date,
        current_date,
        end_date,
        circ_supply_zero,
        locked_fil_zero,
        daily_burnt_fil,
        duration,
        renewal_rate_vec,
        burnt_fil_vec,
        vest_df,
        mint_df,
        known_scheduled_pledge_release_full_vec,
    )
    return cil_df
